import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from qibocal.data import Data, DataUnits
from qibocal.plots.utils import (
    get_color_state0,
    get_color_state1,
    get_data_subfolders,
    load_data,
)


# For calibrate qubit states
def qubit_states(folder, routine, qubit, format):
    figures = []

    fig = make_subplots(
        rows=1,
        cols=1,
        horizontal_spacing=0.1,
        vertical_spacing=0.1,
        # subplot_titles=("Calibrate qubit states"),
    )

    # iterate over multiple data folders
    subfolders = get_data_subfolders(folder)
    report_n = 0
    fitting_report = ""
    max_x, max_y, min_x, min_y = 0, 0, 0, 0
    for subfolder in subfolders:
        try:
            data = load_data(folder, subfolder, routine, format, "data")
            data.df = data.df[data.df["qubit"] == qubit]
        except:
            data = DataUnits(options=["qubit", "iteration", "state"])

        try:
            parameters = load_data(folder, subfolder, routine, format, "parameters")
            parameters.df = parameters.df[parameters.df["qubit"] == qubit]

            average_state0 = complex(parameters.df.iloc[0]["average_state0"])
            average_state1 = complex(
                parameters.df.iloc[0]["average_state1"]  # pylint: disable=E1101
            )
            rotation_angle = parameters.df.iloc[0][  # pylint: disable=E1101
                "rotation_angle"
            ]  # pylint: disable=E1101
            threshold = parameters.df.iloc[0]["threshold"]  # pylint: disable=E1101
            fidelity = parameters.df.iloc[0]["fidelity"]  # pylint: disable=E1101
            assignment_fidelity = parameters.df.iloc[0][
                "assignment_fidelity"
            ]  # pylint: disable=E1101

        except:
            parameters = Data(
                name=f"parameters",
                quantities=[
                    "rotation_angle",  # in rad
                    "threshold",
                    "fidelity",
                    "assignment_fidelity",
                    "average_state0",
                    "average_state1",
                    "qubit",
                ],
            )

        state0_data = data.df[data.df["state"] == 0]
        state1_data = data.df[data.df["state"] == 1]

        fig.add_trace(
            go.Scatter(
                x=state0_data["i"],
                y=state0_data["q"],
                name=f"q{qubit}/r{report_n}: state 0",
                legendgroup=f"q{qubit}/r{report_n}: state 0",
                mode="markers",
                showlegend=False,
                opacity=0.7,
                marker=dict(size=3, color=get_color_state0(report_n)),
            ),
            row=1,
            col=1,
        )

        fig.add_trace(
            go.Scatter(
                x=state1_data["i"],
                y=state1_data["q"],
                name=f"q{qubit}/r{report_n}: state 1",
                legendgroup=f"q{qubit}/r{report_n}: state 1",
                mode="markers",
                showlegend=False,
                opacity=0.7,
                marker=dict(size=3, color=get_color_state1(report_n)),
            ),
            row=1,
            col=1,
        )

        max_x = max(
            max_x,
            state0_data["i"].max(),
            state1_data["i"].max(),
        )
        max_y = max(
            max_y,
            state0_data["q"].max(),
            state1_data["q"].max(),
        )
        min_x = min(
            min_x,
            state0_data["i"].min(),
            state1_data["i"].min(),
        )
        min_y = min(
            min_y,
            state0_data["q"].min(),
            state1_data["q"].min(),
        )

        fig.add_trace(
            go.Scatter(
                x=[average_state0.real],
                y=[average_state0.imag],
                name=f"q{qubit}/r{report_n}: state 0",
                legendgroup=f"q{qubit}/r{report_n}: state 0",
                showlegend=True,
                mode="markers",
                marker=dict(size=10, color=get_color_state0(report_n)),
            ),
            row=1,
            col=1,
        )

        fig.add_trace(
            go.Scatter(
                x=[average_state1.real],
                y=[average_state1.imag],
                name=f"q{qubit}/r{report_n}: state 1",
                legendgroup=f"q{qubit}/r{report_n}: state 1",
                showlegend=True,
                mode="markers",
                marker=dict(size=10, color=get_color_state1(report_n)),
            ),
            row=1,
            col=1,
        )

        title_text = ""
        title_text += (
            f"q{qubit}/r{report_n} | average state 0: ({average_state0:.6f})<br>"
        )
        title_text += (
            f"q{qubit}/r{report_n} | average state 1: ({average_state1:.6f})<br>"
        )
        title_text += f"q{qubit}/r{report_n} | rotation angle: {rotation_angle:.3f}<br>"
        title_text += f"q{qubit}/r{report_n} | threshold: {threshold:.6f}<br>"
        title_text += f"q{qubit}/r{report_n} | fidelity: {fidelity:.3f}<br>"
        title_text += (
            f"q{qubit}/r{report_n} | assignment fidelity: {assignment_fidelity:.3f}<br>"
        )

        fitting_report = fitting_report + title_text
        report_n += 1

    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="i (V)",
        yaxis_title="q (V)",
        xaxis_range=(min_x, max_x),
        yaxis_range=(min_y, max_y),
    )
    fig.update_yaxes(
        scaleanchor="x",
        scaleratio=1,
    )

    figures.append(fig)

    return figures, fitting_report



"""Helper functions for the cli module"""
import datetime
import getpass
import os
import shutil
from glob import glob
from pathlib import Path

import click
import yaml

from qibocal.config import log, raise_error


def folders_exists(folders):
    """Check if a list of folders exists

    Args:
        folders (list): list of absolute or relative path to folders
    """
    foldernames = []
    for foldername in folders:
        expanded = list(glob(foldername))
        if len(expanded) == 0 and "*" not in foldername:
            raise (click.BadParameter(f"file '{foldername}' not found"))
        foldernames.extend(expanded)

    return foldernames


def check_folder_structure(folders):
    """Check if a list of folders share structure between them

    Args:
        folders (list): list of absolute or relative path to folders
    """
    subdirs = [[p for p in Path(folder).glob("**") if p.is_dir()] for folder in folders]
    return all(x == subdirs[0] for x in subdirs)


def update_meta(metadata, metadata_new, target_dir="qq-compare"):
    """Update meta.yml file

    Args:
        metadata (dict): dictionary with the meta.yml actual parameters and values
        metadata_new (dict): dictionary with the new parameters and values to update in the actual meta.yml
    """

    def update(old, new, key):
        old[key] += " , " + new[key]

    for key in ("backend", "date", "end-time", "platform", "start-time", "title"):
        update(metadata, metadata_new, key)
    for key in ("numpy", "qibo", "qibocal", "qibolab"):
        update(metadata["versions"], metadata_new["versions"], key)
    with open(f"{target_dir}/meta.yml", "w") as file:
        yaml.safe_dump(metadata, file)


def update_runcard(rundata, rundata_new, target_compare_dir):
    """Update runcard.yml file

    Args:
        rundata (dict): dictionary with the runcard.yml actual parameters and values
        rundata_new (dict): dictionary with the new parameters and values to update in the actual runcard.yml
    """

    rundata["platform"] += " , " + rundata_new["platform"]
    unique = list(set(rundata["qubits"] + rundata_new["qubits"]))
    rundata["qubits"] = unique
    with open(f"{target_compare_dir}/runcard.yml", "w") as file:
        yaml.safe_dump(
            rundata,
            file,
            indent=4,
            allow_unicode=False,
            sort_keys=False,
            default_flow_style=None,
        )


def load_yaml(path):
    """Load yaml file from disk."""
    with open(path) as file:
        data = yaml.safe_load(file)
    return data


def generate_output_folder(folder, force):
    """Generation of qq output folder

    Args:
        folder (path): path for the output folder. If None it will be created a folder automatically
        force (bool): option to overwrite the output folder if it exists already.

    Returns:
        Output path.
    """
    if folder is None:
        e = datetime.datetime.now()
        user = getpass.getuser().replace(".", "-")
        date = e.strftime("%Y-%m-%d")
        folder = f"{date}-{'000'}-{user}"
        num = 0
        while os.path.exists(folder):
            log.info(f"Directory {folder} already exists.")
            num += 1
            folder = f"{date}-{str(num).rjust(3, '0')}-{user}"
            log.info(f"Trying to create directory {folder}")
    elif os.path.exists(folder) and not force:
        raise_error(RuntimeError, f"Directory {folder} already exists.")
    elif os.path.exists(folder) and force:
        log.warning(f"Deleting previous directory {folder}.")
        shutil.rmtree(os.path.join(os.getcwd(), folder))

    path = os.path.join(os.getcwd(), folder)
    log.info(f"Creating directory {folder}.")
    os.makedirs(path)
    return folder



import os
from colorsys import hls_to_rgb

import numpy as np
import pandas as pd

from qibocal.data import DataUnits


def get_data_subfolders(folder):
    # iterate over multiple data folders
    subfolders = []
    for file in os.listdir(folder):
        d = os.path.join(folder, file)
        if os.path.isdir(d):
            subfolders.append(os.path.basename(d))

    return subfolders[::-1]


def get_color(number):
    return "rgb" + str(hls_to_rgb((0.75 - number * 3 / 20) % 1, 0.4, 0.75))


def get_color_state0(number):
    return "rgb" + str(hls_to_rgb((-0.35 - number * 9 / 20) % 1, 0.6, 0.75))


def get_color_state1(number):
    return "rgb" + str(hls_to_rgb((-0.02 - number * 9 / 20) % 1, 0.6, 0.75))


def load_data(folder, subfolder, routine, data_format, name):
    file = f"{folder}/{subfolder}/{routine}/{name}.csv"
    data = DataUnits()
    all_columns = pd.read_csv(file, nrows=1).columns.tolist()
    if "fit" in name or "parameters" in name:
        data.df = pd.read_csv(file, header=[0], index_col=[0])

    else:
        data.df = pd.read_csv(file, header=[0], skiprows=[1], index_col=[0])

    return data



from enum import Enum

from .allxy.allxy import allxy
from .allxy.allxy_drag_pulse_tuning import allxy_drag_pulse_tuning
from .allxy.drag_pulse_tuning import drag_pulse_tuning
from .classification import single_shot_classification
from .coherence.spin_echo import spin_echo
from .coherence.t1 import t1
from .coherence.t2 import t2
from .dispersive_shift import dispersive_shift
from .flipping import flipping
from .flux_depedence.qubit_flux_dependence import qubit_flux
from .flux_depedence.resonator_flux_dependence import resonator_flux
from .qubit_spectroscopy import qubit_spectroscopy
from .rabi.amplitude import rabi_amplitude
from .rabi.length import rabi_length
from .ramsey import ramsey
from .resonator_punchout import resonator_punchout
from .resonator_punchout_attenuation import resonator_punchout_attenuation
from .resonator_spectroscopy import resonator_spectroscopy


class Operation(Enum):
    resonator_spectroscopy = resonator_spectroscopy
    resonator_punchout = resonator_punchout
    resonator_punchout_attenuation = resonator_punchout_attenuation
    resonator_flux = resonator_flux
    qubit_spectroscopy = qubit_spectroscopy
    qubit_flux = qubit_flux
    rabi_amplitude = rabi_amplitude
    rabi_length = rabi_length
    ramsey = ramsey
    t1 = t1
    t2 = t2
    single_shot_classification = single_shot_classification
    spin_echo = spin_echo
    allxy = allxy
    allxy_drag_pulse_tuning = allxy_drag_pulse_tuning
    drag_pulse_tuning = drag_pulse_tuning
    flipping = flipping
    dispersive_shift = dispersive_shift



from dataclasses import dataclass, field
from typing import Dict, List, Tuple

import numpy as np
import plotly.graph_objects as go
from qibolab.platforms.abstract import AbstractPlatform
from qibolab.pulses import PulseSequence

from qibocal.auto.operation import Parameters, Qubits, Results, Routine
from qibocal.data import DataUnits
from qibocal.plots.utils import get_color_state0, get_color_state1

MESH_SIZE = 50


@dataclass
class SingleShotClassificationParameters(Parameters):
    nshots: int


class SingleShotClassificationData(DataUnits):
    def __init__(self, nshots):
        super().__init__(
            "data",
            options=["qubit", "state"],
        )

        self._nshots = nshots

    @property
    def nshots(self):
        return self._nshots


@dataclass
class SingleShotClassificationResults(Results):
    threshold: Dict[List[Tuple], str] = field(metadata=dict(update="threshold"))
    rotation_angle: Dict[List[Tuple], str] = field(metadata=dict(update="iq_angle"))
    mean_gnd_states: Dict[List[Tuple], str] = field(
        metadata=dict(update="mean_gnd_states")
    )
    mean_exc_states: Dict[List[Tuple], str] = field(
        metadata=dict(update="mean_exc_states")
    )
    fidelity: Dict[List[Tuple], str]
    assignment_fidelity: Dict[List[Tuple], str]


def _acquisition(
    params: SingleShotClassificationParameters,
    platform: AbstractPlatform,
    qubits: Qubits,
) -> SingleShotClassificationData:
    """
    Method which implements the state's calibration of a chosen qubit. Two analogous tests are performed
    for calibrate the ground state and the excited state of the oscillator.
    The subscripts `exc` and `gnd` will represent the excited state |1> and the ground state |0>.

    Args:
        platform (:class:`qibolab.platforms.abstract.AbstractPlatform`): custom abstract platform on which we perform the calibration.
        qubits (dict): Dict of target Qubit objects to perform the action
        nshots (int): number of times the pulse sequence will be repeated.

    Returns:
        A DataUnits object with the raw data obtained for the fast and precision sweeps with the following keys

            - **MSR[V]**: Resonator signal voltage mesurement in volts
            - **i[V]**: Resonator signal voltage mesurement for the component I in volts
            - **q[V]**: Resonator signal voltage mesurement for the component Q in volts
            - **phase[rad]**: Resonator signal phase mesurement in radians
            - **iteration[dimensionless]**: Execution number
            - **qubit**: The qubit being tested
            - **iteration**: The iteration number of the many determined by software_averages

    """

    # create two sequences of pulses:
    # state0_sequence: I  - MZ
    # state1_sequence: RX - MZ

    # taking advantage of multiplexing, apply the same set of gates to all qubits in parallel
    state0_sequence = PulseSequence()
    state1_sequence = PulseSequence()

    RX_pulses = {}
    ro_pulses = {}
    for qubit in qubits:
        RX_pulses[qubit] = platform.create_RX_pulse(qubit, start=0)
        ro_pulses[qubit] = platform.create_qubit_readout_pulse(
            qubit, start=RX_pulses[qubit].finish
        )

        state0_sequence.add(ro_pulses[qubit])
        state1_sequence.add(RX_pulses[qubit])
        state1_sequence.add(ro_pulses[qubit])

    # create a DataUnits object to store the results
    data = SingleShotClassificationData(params.nshots)

    # execute the first pulse sequence
    state0_results = platform.execute_pulse_sequence(
        state0_sequence, nshots=params.nshots
    )

    # retrieve and store the results for every qubit
    for ro_pulse in ro_pulses.values():
        r = state0_results[ro_pulse.serial].raw
        r.update(
            {
                "qubit": [ro_pulse.qubit] * params.nshots,
                "state": [0] * params.nshots,
            }
        )
        data.add_data_from_dict(r)

    # execute the second pulse sequence
    state1_results = platform.execute_pulse_sequence(
        state1_sequence, nshots=params.nshots
    )

    # retrieve and store the results for every qubit
    for ro_pulse in ro_pulses.values():
        r = state1_results[ro_pulse.serial].raw
        r.update(
            {
                "qubit": [ro_pulse.qubit] * params.nshots,
                "state": [1] * params.nshots,
            }
        )
        data.add_data_from_dict(r)

    return data


def _fit(data: SingleShotClassificationData) -> SingleShotClassificationResults:
    qubits = data.df["qubit"].unique()
    thresholds, rotation_angles = {}, {}
    fidelities, assignment_fidelities = {}, {}
    mean_gnd_states = {}
    mean_exc_states = {}

    for qubit in qubits:
        qubit_data = data.df[data.df["qubit"] == qubit].drop(
            columns=["qubit", "MSR", "phase"]
        )

        iq_state0 = (
            qubit_data[qubit_data["state"] == 0]["i"].pint.to("V").pint.magnitude
            + 1.0j
            * qubit_data[qubit_data["state"] == 0]["q"].pint.to("V").pint.magnitude
        )
        iq_state1 = (
            qubit_data[qubit_data["state"] == 1]["i"].pint.to("V").pint.magnitude
            + 1.0j
            * qubit_data[qubit_data["state"] == 1]["q"].pint.to("V").pint.magnitude
        )

        iq_state1 = np.array(iq_state1)
        iq_state0 = np.array(iq_state0)

        iq_mean_state1 = np.mean(iq_state1)
        iq_mean_state0 = np.mean(iq_state0)

        vector01 = iq_mean_state1 - iq_mean_state0
        rotation_angle = np.angle(vector01)

        iq_state1_rotated = iq_state1 * np.exp(-1j * rotation_angle)
        iq_state0_rotated = iq_state0 * np.exp(-1j * rotation_angle)

        real_values_state1 = iq_state1_rotated.real
        real_values_state0 = iq_state0_rotated.real

        real_values_combined = np.concatenate((real_values_state1, real_values_state0))
        real_values_combined.sort()

        cum_distribution_state1 = [
            sum(map(lambda x: x.real >= real_value, real_values_state1))
            for real_value in real_values_combined
        ]
        cum_distribution_state0 = [
            sum(map(lambda x: x.real >= real_value, real_values_state0))
            for real_value in real_values_combined
        ]

        cum_distribution_diff = np.abs(
            np.array(cum_distribution_state1) - np.array(cum_distribution_state0)
        )
        argmax = np.argmax(cum_distribution_diff)
        threshold = real_values_combined[argmax]
        errors_state1 = data.nshots - cum_distribution_state1[argmax]
        errors_state0 = cum_distribution_state0[argmax]
        fidelity = cum_distribution_diff[argmax] / data.nshots
        assignment_fidelity = 1 - (errors_state1 + errors_state0) / data.nshots / 2
        thresholds[qubit] = threshold
        rotation_angles[
            qubit
        ] = -rotation_angle  # TODO: qblox driver np.rad2deg(-rotation_angle)
        fidelities[qubit] = fidelity
        mean_gnd_states[qubit] = iq_mean_state0
        mean_exc_states[qubit] = iq_mean_state1
        assignment_fidelities[qubit] = assignment_fidelity

    return SingleShotClassificationResults(
        thresholds,
        rotation_angles,
        mean_gnd_states,
        mean_exc_states,
        fidelities,
        assignment_fidelities,
    )


def _plot(
    data: SingleShotClassificationData, fit: SingleShotClassificationResults, qubit
):
    figures = []

    fig = go.Figure()

    fitting_report = ""
    max_x, max_y, min_x, min_y = 0, 0, 0, 0

    qubit_data = data.df[data.df["qubit"] == qubit]
    state0_data = qubit_data[data.df["state"] == 0].drop(
        columns=["MSR", "phase", "qubit"]
    )
    state1_data = qubit_data[data.df["state"] == 1].drop(
        columns=["MSR", "phase", "qubit"]
    )

    fig.add_trace(
        go.Scatter(
            x=state0_data["i"].pint.to("V").pint.magnitude,
            y=state0_data["q"].pint.to("V").pint.magnitude,
            name="Ground State",
            legendgroup="Ground State",
            mode="markers",
            showlegend=True,
            opacity=0.7,
            marker=dict(size=3, color=get_color_state0(0)),
        ),
    )

    fig.add_trace(
        go.Scatter(
            x=state1_data["i"].pint.to("V").pint.magnitude,
            y=state1_data["q"].pint.to("V").pint.magnitude,
            name="Excited State",
            legendgroup="Excited State",
            mode="markers",
            showlegend=True,
            opacity=0.7,
            marker=dict(size=3, color=get_color_state1(0)),
        ),
    )

    fig.add_trace(
        go.Scatter(
            x=[state0_data["i"].pint.to("V").pint.magnitude.mean()],
            y=[state0_data["q"].pint.to("V").pint.magnitude.mean()],
            name="Average Ground State",
            legendgroup="Average Ground State",
            showlegend=True,
            mode="markers",
            marker=dict(size=10, color=get_color_state0(0)),
        ),
    )

    fig.add_trace(
        go.Scatter(
            x=[state1_data["i"].pint.to("V").pint.magnitude.mean()],
            y=[state1_data["q"].pint.to("V").pint.magnitude.mean()],
            name="Average Excited State",
            legendgroup="Average Excited State",
            showlegend=True,
            mode="markers",
            marker=dict(size=10, color=get_color_state1(0)),
        ),
    )

    max_x = max(
        max_x,
        state0_data["i"].pint.to("V").pint.magnitude.max(),
        state1_data["i"].pint.to("V").pint.magnitude.max(),
    )
    max_y = max(
        max_y,
        state0_data["q"].pint.to("V").pint.magnitude.max(),
        state1_data["q"].pint.to("V").pint.magnitude.max(),
    )
    min_x = min(
        min_x,
        state0_data["i"].pint.to("V").pint.magnitude.min(),
        state1_data["i"].pint.to("V").pint.magnitude.min(),
    )
    min_y = min(
        min_y,
        state0_data["q"].pint.to("V").pint.magnitude.min(),
        state1_data["q"].pint.to("V").pint.magnitude.min(),
    )

    feature_x = np.linspace(min_x, max_x, MESH_SIZE)
    feature_y = np.linspace(min_y, max_y, MESH_SIZE)

    [x, y] = np.meshgrid(feature_x, feature_y)

    z = (
        (np.exp(1j * fit.rotation_angle[qubit]) * (x + 1j * y)).real
        > fit.threshold[qubit]
    ).astype(int)
    fig.add_trace(
        go.Contour(
            x=feature_x,
            y=feature_y,
            z=z,
            showscale=False,
            colorscale=[get_color_state0(0), get_color_state1(0)],
            opacity=0.4,
            name="Score",
            hoverinfo="skip",
        ),
    )

    fitting_report = (
        fitting_report
        + f"{qubit} | Average Ground State: {fit.mean_gnd_states[qubit]:.4f} <br>"
        + f"{qubit} | Average Excited State: {fit.mean_exc_states[qubit]:.4f} <br>"
        + f"{qubit} | Rotation Angle: {fit.rotation_angle[qubit]:.3f} rad <br>"
        + f"{qubit} | Threshold: {fit.threshold[qubit]:.4f} <br>"
        + f"{qubit} | Fidelity: {fit.fidelity[qubit]:.3f} <br>"
        + f"{qubit} | Assignment Fidelity: {fit.assignment_fidelity[qubit]:.3f} <br>"
    )

    fig.update_layout(
        showlegend=True,
        uirevision="0",  # ``uirevision`` allows zooming while live plotting
        xaxis_title="i (V)",
        yaxis_title="q (V)",
        xaxis_range=(min_x, max_x),
        yaxis_range=(min_y, max_y),
    )
    fig.update_yaxes(
        scaleanchor="x",
        scaleratio=1,
    )

    figures.append(fig)

    return figures, fitting_report


single_shot_classification = Routine(_acquisition, _fit, _plot)



""" This script implements an easy 1-qubit protocol with only X-gates and identities.
It is a great example on how to write an own niGSC protocol. The single functions above have
little descriptions for the purpose of that function and what is important to include.

1. Step:

  Define the two module specific classes which are used in defining and executing an experiment,
  the circuit factory and experiment class.
  They can also just inherit everything from another module.

2. Step:

  Write the sequential post processing functions.

3. Step:

  Write the aggregational post processing function.

4. Step:

  Write the function to build a report. When using the qq module, a plotly figure has to be returned.
"""

# These libraries should be enough when starting a new protocol.
from __future__ import annotations

from collections.abc import Iterable

import numpy as np
import pandas as pd
from plotly.graph_objects import Figure
from qibo import gates
from qibo.models import Circuit
from qibo.noise import NoiseModel

import qibocal.calibrations.niGSC.basics.fitting as fitting_methods
from qibocal.calibrations.niGSC.basics.circuitfactory import CircuitFactory
from qibocal.calibrations.niGSC.basics.experiment import Experiment
from qibocal.calibrations.niGSC.basics.plot import Report, scatter_fit_fig
from qibocal.calibrations.niGSC.basics.utils import number_to_str
from qibocal.config import raise_error


# Define the circuit factory class for this specific module.
class ModuleFactory(CircuitFactory):
    def __init__(self, nqubits: int, depths: list, qubits: list = []) -> None:
        super().__init__(nqubits, depths, qubits)
        if not len(self.qubits) == 1:
            raise_error(
                ValueError,
                f"This class is written for gates acting on only one qubit, not {len(self.qubits)} qubits.",
            )
        self.name = "XId"

    def build_circuit(self, depth: int):
        # Initiate the empty circuit from qibo with 'self.nqubits'
        # many qubits.
        circuit = Circuit(1)
        # Draw sequence length many I and X gates.
        gate_lists = np.random.choice([gates.I(0), gates.X(0)], size=depth)
        # Add gates to circuit.
        circuit.add(gate_lists)
        circuit.add(gates.M(0))
        return circuit


# Define the experiment class for this specific module.
class ModuleExperiment(Experiment):
    def __init__(
        self,
        circuitfactory: Iterable,
        data: Iterable | None = None,
        nshots: int | None = None,
        noise_model: NoiseModel = None,
    ) -> None:
        super().__init__(circuitfactory, data, nshots, noise_model)
        # Make the circuitfactory a list be able to store them using save_circuits method.
        self.prebuild()
        self.name = "XIdRB"

    def execute(self, circuit: Circuit, datarow: dict) -> dict:
        datadict = super().execute(circuit, datarow)
        datadict["depth"] = circuit.ngates - 1
        # TODO change that.
        datadict["countX"] = len(circuit.gates_of_type("x"))
        return datadict


# Define the result class for this specific module.
class ModuleReport(Report):
    def __init__(self) -> None:
        super().__init__()
        self.title = "X-Id Benchmarking"


# The filter functions/post processing functions always dependent on circuit and data row!
# It is executed row by row when used on an experiment object.
def filter_sign(circuit: Circuit, datarow: dict) -> dict:
    """Calculates the filtered signal for the XId.

    :math:`n_X` denotes the amount of :math:`X` gates in the circuit with gates
    :math:`g` and :math`i` the outcome which is either ground state :math:`0`
    or exited state :math:`1`.

    .. math::
        f_{\\text{sign}}(i,g)
        = (-1)^{n_X\\%2 + i}/2


    Args:
        circuit (Circuit): Not needed here.
        datarow (dict): The dictionary with the samples from executed circuits and amount of
                        X gates in the executed circuit.

    Returns:
        dict: _description_
    """
    samples = datarow["samples"]
    countX = datarow["countX"]
    filtersign = 0
    for s in samples:
        filtersign += (-1) ** (countX % 2 + s[0]) / 2.0
    datarow["filter"] = filtersign / len(samples)
    return datarow


# All the predefined sequential postprocessing / filter functions are bundled together here.
def post_processing_sequential(experiment: Experiment):
    """Perform sequential tasks needed to analyze the experiment results.

    The data is added/changed in the experiment, nothing has to be returned.

    Args:
        experiment (Experiment): Experiment object after execution of the experiment itself.
    """

    # Compute and add the ground state probabilities row by row.
    experiment.perform(filter_sign)


# After the row by row execution of tasks comes the aggregational task. Something like calculation
# of means, deviations, fitting data, tasks where the whole data as to be looked at, and not just
# one instance of circuit + other information.
def get_aggregational_data(experiment: Experiment, ndecays: int = 2) -> pd.DataFrame:
    """Computes aggregational tasks, fits data and stores the results in a data frame.

    No data is manipulated in the ``experiment`` object.

    Args:
        experiment (Experiment): After sequential postprocessing of the experiment data.
        ndecays (int): Number of decays to be fitted. Default is 2.

    Returns:
        pd.DataFrame: The summarized data.
    """
    # Has to fit the column describtion from ``filter_sign``.
    depths, ydata = experiment.extract("filter", "depth", "mean")
    _, ydata_std = experiment.extract("filter", "depth", "std")
    # Fit the filtered signal for each depth, there could be two overlaying exponential functions.
    popt, perr = fitting_methods.fit_expn_func(depths, ydata, n=ndecays)
    # Create dictionaries with fitting parameters and estimated errors
    popt_keys = [f"A{k+1}" for k in range(ndecays)]
    popt_keys += [f"p{k+1}" for k in range(ndecays)]
    popt_dict = dict(zip(popt_keys, popt))
    perr_keys = [f"A{k+1}_err" for k in range(ndecays)]
    perr_keys += [f"p{k+1}_err" for k in range(ndecays)]
    perr_dict = dict(zip(perr_keys, perr))
    # Build a list of dictionaries with the aggregational information.
    data = [
        {
            "depth": depths,  # The x-axis.
            "data": ydata,  # The filtred signal.
            "2sigma": 2 * ydata_std,  # The 2 * standard deviation error for each depth.
            "fit_func": "expn_func",  # Which function was used to fit.
            "popt": popt_dict,  # The real fitting parameters.
            "perr": perr_dict,  # The estimated errors.
        }
    ]

    df = pd.DataFrame(data, index=["filter"])
    return df


# This is highly individual. The only important thing for the qq module is that a plotly figure is
# returned, if qq is not used any type of figure can be build.
def build_report(experiment: Experiment, df_aggr: pd.DataFrame) -> Figure:
    """Use data and information from ``experiment`` and the aggregated data data frame to
    build a report as plotly figure.

    Args:
        experiment (Experiment): After sequential postprocessing of the experiment data.
        df_aggr (pd.DataFrame): Normally build with ``get_aggregational_data`` function.

    Returns:
        (tuple): A `plotly.graphical_object.Figure` object and a string corresponding to a table.
    """

    # Initiate a report object.
    report = ModuleReport()
    # Add general information to the table.
    report.info_dict["Number of qubits"] = len(experiment.data[0]["samples"][0])
    report.info_dict["Number of shots"] = len(experiment.data[0]["samples"])
    report.info_dict["runs"] = experiment.extract("samples", "depth", "count")[1][0]
    dfrow = df_aggr.loc["filter"]
    popt_pairs = list(dfrow["popt"].items())[::2] + list(dfrow["popt"].items())[1::2]
    report.info_dict["Fit"] = "".join(
        [f"{key}={number_to_str(value)} " for key, value in popt_pairs]
    )
    perr_pairs = list(dfrow["perr"].items())[::2] + list(dfrow["perr"].items())[1::2]
    report.info_dict["Fitting deviations"] = "".join(
        [f"{key}={number_to_str(value)} " for key, value in perr_pairs]
    )
    # Use the predefined ``scatter_fit_fig`` function from ``basics.utils`` to build the wanted
    # plotly figure with the scattered filtered data along with the mean for
    # each depth and the exponential fit for the means.
    report.all_figures.append(scatter_fit_fig(experiment, df_aggr, "depth", "filter"))

    # Return the figure of the report object and the corresponding table.
    return report.build()



import numpy as np
from scipy.optimize import curve_fit

from qibocal.config import log


def exp_decay(x, *p):
    return p[0] - p[1] * np.exp(-1 * x * p[2])


def exponential_fit(data):
    qubits = data.df["qubit"].unique()

    decay = {}
    fitted_parameters = {}

    for qubit in qubits:
        qubit_data_df = data.df[data.df["qubit"] == qubit]
        voltages = qubit_data_df["MSR"].pint.to("uV").pint.magnitude
        times = qubit_data_df["wait"].pint.to("ns").pint.magnitude

        try:
            y_max = np.max(voltages.values)
            y_min = np.min(voltages.values)
            y = (voltages.values - y_min) / (y_max - y_min)
            x_max = np.max(times.values)
            x_min = np.min(times.values)
            x = (times.values - x_min) / (x_max - x_min)

            p0 = [
                0.5,
                0.5,
                5,
            ]
            popt = curve_fit(
                exp_decay,
                x,
                y,
                p0=p0,
                maxfev=2000000,
                bounds=(
                    [-2, -2, 0],
                    [2, 2, np.inf],
                ),
            )[0]
            popt = [
                (y_max - y_min) * popt[0] + y_min,
                (y_max - y_min) * popt[1] * np.exp(x_min * popt[2] / (x_max - x_min)),
                popt[2] / (x_max - x_min),
            ]
            t2 = 1.0 / popt[2]

        except Exception as e:
            log.warning(f"Exp decay fitting was not succesful. {e}")
            popt = [0] * 3
            t2 = 5.0

        fitted_parameters[qubit] = popt
        decay[qubit] = t2

    return decay, fitted_parameters



"""Adds global CLI options."""
import base64
import os
import pathlib
import shutil
import socket
import subprocess
import uuid
from datetime import datetime
from glob import glob
from urllib.parse import urljoin

import click
from qibo.config import log, raise_error

from ..cli.auto_builder import AutoCalibrationBuilder
from ..cli.builders import ActionBuilder, load_yaml
from .utils import check_folder_structure, folders_exists, update_meta, update_runcard

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])

# options for report upload
UPLOAD_HOST = (
    "qibocal@localhost"
    if socket.gethostname() == "saadiyat"
    else "qibocal@login.qrccluster.com"
)
TARGET_DIR = "qibocal-reports/"
ROOT_URL = "http://login.qrccluster.com:9000/"

TARGET_COMPARE_DIR = "qq-compare/"


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("runcard", metavar="RUNCARD", type=click.Path(exists=True))
@click.option(
    "folder",
    "-o",
    type=click.Path(),
    help="Output folder. If not provided a standard name will generated.",
)
@click.option(
    "force",
    "-f",
    is_flag=True,
    help="Use --force option to overwrite the output folder.",
)
def command(runcard, folder, force=None):
    """qibocal: Quantum Calibration Verification and Validation using Qibo.

    Arguments:

     - RUNCARD: runcard with declarative inputs.

     - PLATFORM_RUNCARD: Qibolab's platform runcard. If not provided Qibocal will use the runcard available in Qibolab for the platform chosen.
    """

    builder = ActionBuilder(runcard, folder, force)
    builder.execute()


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("runcard", metavar="RUNCARD", type=click.Path(exists=True))
@click.option(
    "folder",
    "-o",
    type=click.Path(),
    help="Output folder. If not provided a standard name will generated.",
)
@click.option(
    "force",
    "-f",
    is_flag=True,
    help="Use --force option to overwrite the output folder.",
)
def autocalibration(runcard, folder, force=None):
    """qibocal: Quantum Calibration Verification and Validation using Qibo.

    Arguments:

     - RUNCARD: runcard with declarative inputs.

     - PLATFORM_RUNCARD: Qibolab's platform runcard. If not provided Qibocal will use the runcard available in Qibolab for the platform chosen.
    """

    builder = AutoCalibrationBuilder(runcard, folder, force)
    builder.run()
    builder.dump_platform_runcard()
    builder.dump_report()


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "port",
    "-p",
    "--port",
    default=8050,
    type=int,
    help="Localhost port to launch dash server.",
)
@click.option(
    "debug",
    "-d",
    "--debug",
    is_flag=True,
    help="Launch server in debugging mode.",
)
def live_plot(port, debug):
    """Real time plotting of calibration data on a dash server."""
    import socket

    from qibocal.web.app import app

    # change port if it is already used
    while True:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            if s.connect_ex(("localhost", port)) != 0:
                break
        port += 1

    app.run_server(debug=debug, port=port)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("output_folder", metavar="FOLDER", type=click.Path(exists=True))
def upload(output_folder):
    """Uploads output folder to server"""

    output_path = pathlib.Path(output_folder)

    # check the rsync command exists.
    if not shutil.which("rsync"):
        raise_error(
            RuntimeError,
            "Could not find the rsync command. Please make sure it is installed.",
        )

    # check that we can authentica with a certificate
    ssh_command_line = (
        "ssh",
        "-o",
        "PreferredAuthentications=publickey",
        "-q",
        UPLOAD_HOST,
        "exit",
    )

    str_line = " ".join(repr(ele) for ele in ssh_command_line)

    log.info(f"Checking SSH connection to {UPLOAD_HOST}.")

    try:
        subprocess.run(ssh_command_line, check=True)
    except subprocess.CalledProcessError as e:
        raise RuntimeError(
            (
                "Could not validate the SSH key. "
                "The command\n%s\nreturned a non zero exit status. "
                "Please make sure that your public SSH key is on the server."
            )
            % str_line
        ) from e
    except OSError as e:
        raise RuntimeError(
            "Could not run the command\n{}\n: {}".format(str_line, e)
        ) from e

    log.info("Connection seems OK.")

    # upload output
    randname = base64.urlsafe_b64encode(uuid.uuid4().bytes).decode()
    newdir = TARGET_DIR + randname

    rsync_command = (
        "rsync",
        "-aLz",
        "--chmod=ug=rwx,o=rx",
        f"{output_path}/",
        f"{UPLOAD_HOST}:{newdir}",
    )

    log.info(f"Uploading output ({output_path}) to {UPLOAD_HOST}")
    try:
        subprocess.run(rsync_command, check=True)
    except subprocess.CalledProcessError as e:
        msg = f"Failed to upload output: {e}"
        raise RuntimeError(msg) from e

    url = urljoin(ROOT_URL, randname)
    log.info(f"Upload completed. The result is available at:\n{url}")


# qq-compare
@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("folders", metavar="FOLDER", type=click.Path(exists=True), nargs=-1)
def compare(folders):
    """Creates a comparision folder given N different data folders following qq-live data structure to be visualized

    Args:
        folder1 (string): absolute or relative path of data folder 1 to be compared
        ...
        folderN (string): absolute or relative path of data folder N to be compared
    """

    # check list of folders exists.
    foldernames = folders_exists(folders)

    # check folder structures matches
    if not check_folder_structure(foldernames):
        raise_error(
            RuntimeError,
            "Could not compare the list of folders. Please make sure that they have the same structure",
        )
    else:
        log.info(f"Folders are comparable.")

    # move old compare folder and remove
    now = str(datetime.today())
    tmp_folder = "qq-compare_" + now.replace(" ", "_")
    if os.path.isdir(TARGET_COMPARE_DIR):
        os.rename(TARGET_COMPARE_DIR, tmp_folder)

    os.mkdir(TARGET_COMPARE_DIR)
    # TODO: prepare meta.yml and runcard.yml as mix of all meta and runcard files
    shutil.copy(pathlib.Path(foldernames[0]).joinpath("meta.yml"), TARGET_COMPARE_DIR)
    shutil.copy(
        pathlib.Path(foldernames[0]).joinpath("runcard.yml"), TARGET_COMPARE_DIR
    )

    for i, folder in enumerate(foldernames):
        newdir = TARGET_COMPARE_DIR + f"data{i}"
        log.info(f"Copying ({folder}) into {newdir}")
        try:
            shutil.copytree(pathlib.Path(folder).joinpath("data"), newdir)
            for file in glob(os.path.join(folder, "*.html")):
                shutil.copy2(file, newdir)
            for file in glob(os.path.join(folder, "*.yml")):
                shutil.copy2(file, newdir)
        except subprocess.CalledProcessError as e:
            msg = f"Failed to upload output: {e}"
            raise RuntimeError(msg) from e

        if i > 0:
            # update meta.yml for comparing report
            metadata_new = load_yaml(os.path.join(folder, "meta.yml"))
            metadata = load_yaml(os.path.join(TARGET_COMPARE_DIR, "meta.yml"))
            update_meta(metadata, metadata_new)
            # update runcard.yml for comparing report
            rundata_new = load_yaml(os.path.join(folder, "runcard.yml"))
            rundata = load_yaml(os.path.join(TARGET_COMPARE_DIR, "runcard.yml"))
            update_runcard(rundata, rundata_new, TARGET_COMPARE_DIR)

    log.info(f"Upload completed")

    from qibocal.web.report import create_report

    create_report(TARGET_COMPARE_DIR)
    log.info(f"HTML report generated")



import pathlib

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from sklearn.model_selection import train_test_split

COLORS = ["#FF0000", "#0000FF"]
DATAFILE = "data_processing.pdf"


def load_qubit(data_path: pathlib.Path, qubit):
    r"""Load the information of the qubit `qubit`
    stored in `data_path`.
    Args:
        data_path (path): Data path.
        qubit: Qubit ID.
    Returns:
        Pandas.DataFrame with `qubit`'s data.
    """
    data = pd.read_csv(data_path, skiprows=[1])
    data = data[data.qubit == qubit]
    return data


def generate_models(data, test_size=0.25):
    r"""Extract from data the values stored
    in the keys `i` and `q` and split them in training and test sets.

    Args:
        data (DataFrame): Qubit's info.
        test_size (float | Optional): The proportion of the dataset to include in the train split.

    Returns:

        - x_train: Training inputs.
        - y_train: Training outputs.
        - x_test: Test inputs.
        - y_test: Test outputs.
    """
    input_data = data[["i", "q"]].values
    output_data = data["state"].values
    return train_test_split(
        input_data, output_data, test_size=test_size, random_state=0, shuffle=True
    )


def plot_qubit(data, save_dir: pathlib.Path):
    r"""Plot the `data` and save it in `{save_dir}/data_processing.pdf`.

    Args:
        data (DataFrame): Input data with "i", "q" and "state" keys.
        save_dir (path): Save path.
    """
    _, axes = plt.subplots(1, 2, figsize=(14, 7))

    sns.set_palette(sns.color_palette(COLORS))
    sns.scatterplot(
        x="i", y="q", data=data, hue="state", ax=axes[0], alpha=0.7, edgecolor="black"
    )
    qubit = int(data.iloc[0]["qubit"])
    axes[0].set_title(f"qubit {qubit}")
    sns.countplot(x=data.state, data=data, ax=axes[1])
    axes[1].set_title("states distribution")
    plt.tight_layout()
    plt.savefig(save_dir / DATAFILE)
